from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from io import BytesIO
from . import core
from . import utils
from .core import AVBPropertyDef, AVBRefList
from . utils import peek_data

@utils.register_class
class MediaDescriptor(core.AVBObject):
    class_id = b'MDES'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('mob_kind',       'OMFI:MDES:MobKind',         'int8',         0),
        AVBPropertyDef('locator',        'OMFI:MDES:Locator',         'reference', None),
        AVBPropertyDef('intermediate',   'OMFI:MDES:MC:Intermediate', 'bool',     False),
        AVBPropertyDef('physical_media', 'OMFI:MOBJ:PhysicalMedia',   'reference', None),
        AVBPropertyDef('uuid',           'OMFI:AMDL:acfUID',          'UUID'),
        AVBPropertyDef('wchar',          'OMFI:AMDL:acfWChar',        'bytes'),
        AVBPropertyDef('attributes',     'OMFI:AMDL:Attributes',      'reference'),
    ]
    __slots__ = ()

    def read(self, f):
        super(MediaDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        self.mob_kind = ctx.read_u8(f)
        self.locator = []
        self.locator = ctx.read_object_ref(self.root, f)
        self.intermediate = ctx.read_bool(f)
        self.physical_media = ctx.read_object_ref(self.root, f)

        # print('sss', self.locator)
        # print(peek_data(f).encode('hex'))

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 65)
                uuid_len = ctx.read_s32(f)
                assert uuid_len == 16
                self.uuid = ctx.read_raw_uuid(f)
            elif tag == 0x02:
                # this is utf-16 string data
                ctx.read_assert_tag(f, 65)
                size = ctx.read_s32(f)
                self.wchar = bytearray(f.read(size))
            elif tag == 0x03:
                ctx.read_assert_tag(f, 72)
                self.attributes = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] == b'MDES':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MediaDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_u8(f, self.mob_kind)
        ctx.write_object_ref(self.root, f, self.locator)
        ctx.write_bool(f, self.intermediate)
        ctx.write_object_ref(self.root, f, self.physical_media)

        if hasattr(self, 'uuid'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 65)
            ctx.write_s32(f, 16)
            ctx.write_raw_uuid(f, self.uuid)

        if hasattr(self, 'wchar'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 65)
            ctx.write_s32(f, len(self.wchar))
            f.write(self.wchar)

        if hasattr(self, 'attributes'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.attributes)

        if self.class_id[:] == b'MDES':
            ctx.write_u8(f, 0x03)


@utils.register_class
class TapeDescriptor(MediaDescriptor):
    class_id = b'MDTP'
    propertydefs = MediaDescriptor.propertydefs + [
        AVBPropertyDef('cframe', 'OMFI:MDTP:CFrame', "int16",  0)
    ]
    __slots__ = ()

    def read(self, f):
        super(TapeDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        self.cframe = ctx.read_s16(f)
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TapeDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        ctx.write_s16(f, self.cframe)
        ctx.write_u8(f, 0x03)

@utils.register_class
class FilmDescriptor(MediaDescriptor):
    class_id = b'MDFM'
    __slots__ = ()

    def read(self, f):
        super(FilmDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(FilmDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)
        ctx.write_u8(f, 0x03)

@utils.register_class
class NagraDescriptor(MediaDescriptor):
    class_id = b'MDNG'
    __slots__ = ()

    def read(self, f):
        super(NagraDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(NagraDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)
        ctx.write_u8(f, 0x03)

@utils.register_class
class MediaFileDescriptor(MediaDescriptor):
    class_id = b'MDFL'
    propertydefs = MediaDescriptor.propertydefs + [
        AVBPropertyDef('edit_rate',   'EdRate',               'fexp10', 25),
        AVBPropertyDef('length',      'OMFI:MDFL:Length',     'int32',   0),
        AVBPropertyDef('is_omfi',     'OMFI:MDFL:IsOMFI',     'int16',   0),
        AVBPropertyDef('data_offset', 'OMFI:MDFL:dataOffset', 'int32',   0),
    ]
    __slots__ = ()

    def read(self, f):
        super(MediaFileDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        self.edit_rate = ctx.read_exp10_encoded_float(f)
        self.length = ctx.read_s32(f)
        self.is_omfi = ctx.read_s16(f)
        self.data_offset = ctx.read_s32(f)

        if self.class_id[:] == b'MDFL':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MediaFileDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_exp10_encoded_float(f, self.edit_rate)
        ctx.write_s32(f, self.length)
        ctx.write_s16(f, self.is_omfi)
        ctx.write_s32(f, self.data_offset)

        if self.class_id[:] == b'MDFL':
            ctx.write_u8(f, 0x03)

@utils.register_class
class MultiDescriptor(MediaFileDescriptor):
    class_id = b'MULD'
    propertydefs = MediaFileDescriptor.propertydefs + [
        AVBPropertyDef('descriptors', 'OMFI:MULD:Descriptors', "ref_list"),
    ]
    __slots__ = ()

    def read(self, f):
        super(MultiDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        count = ctx.read_s32(f)
        self.descriptors = AVBRefList.__new__(AVBRefList, root=self.root)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.descriptors.append(ref)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MultiDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self.descriptors))

        for descriptor in self.descriptors:
            ctx.write_object_ref(self.root, f, descriptor)

        ctx.write_u8(f, 0x03)

@utils.register_class
class WaveDescriptor(MediaFileDescriptor):
    class_id = b'WAVE'
    propertydefs = MediaFileDescriptor.propertydefs + [
        AVBPropertyDef('summary',   'OMFI:WAVD:Summary',   'bytes'),
    ]
    __slots__ = ()

    def read(self, f):
        super(WaveDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        assert f.read(4) == b'RIFF'

        # NOTE: this is suppose to be LE
        size = ctx.read_u32le(f)
        self.summary = bytearray(f.read(size))
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(WaveDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)
        f.write(b'RIFF')
        # NOTE: this is suppose to be LE
        ctx.write_u32le(f, len(self.summary))
        f.write(self.summary)
        ctx.write_u8(f, 0x03)

@utils.register_class
class AIFCDescriptor(MediaFileDescriptor):
    class_id = b'AIFC'
    propertydefs = MediaFileDescriptor.propertydefs + [
        AVBPropertyDef('summary',   'OMFI:AIFD:Summary',   'bytes'),
        AVBPropertyDef('data_pos',  'OMFI:AIFD:MC:DataPos', 'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(AIFCDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        assert f.read(4) == b'FORM'

        # NOTE: this is suppose to be BE
        size = ctx.read_u32be(f)
        self.summary = bytearray(f.read(size))
        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.data_pos = ctx.read_s32(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(AIFCDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        f.write(b'FORM')
        # NOTE: this is suppose to be BE
        ctx.write_u32be(f, len(self.summary))
        f.write(self.summary)

        if hasattr(self, 'data_pos'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.data_pos)

        ctx.write_u8(f, 0x03)

@utils.register_class
class PCMADescriptor(MediaFileDescriptor):
    class_id = b'PCMA'
    propertydefs = MediaFileDescriptor.propertydefs + [
        AVBPropertyDef('channels',                    'OMFI:MDAU:NumChannels',               'uint16'),
        AVBPropertyDef('quantization_bits',           'OMFI:MDAU:BitsPerSample',             'uint16'),
        AVBPropertyDef('sample_rate',                 'EdRate',                              'fexp10'),
        AVBPropertyDef('locked',                      'OMFI:PCMA:Locked',                    'bool'),
        AVBPropertyDef('audio_ref_level',             'OMFI:PCMA:AudioRefLevel',             'int16'),
        AVBPropertyDef('electro_spatial_formulation', 'OMFI:PCMA:ElectroSpatialFormulation', 'int32'),
        AVBPropertyDef('dial_norm',                   'OMFI:PCMA:DialNorm',                  'uint16'),
        AVBPropertyDef('coding_format',               'OMFI:PCMA:AudioCodingFormat',         'int32'),
        AVBPropertyDef('block_align',                 'OMFI:PCMA:BlockAlignment',            'int32'),
        AVBPropertyDef('sequence_offset',             'OMFI:PCMA:SequenceOffset',            'uint16'),
        AVBPropertyDef('average_bps',                 'OMFI:PCMA:AverageBytesPerSecond',     'int32'),
        AVBPropertyDef('has_peak_envelope_data',      'OMFI:PCMA:HasPeakEnvelopeData',       'bool'),
        AVBPropertyDef('peak_envelope_version',       'OMFI:PCMA:PeakEnvelopeVersion',       'int32'),
        AVBPropertyDef('peak_envelope_format',        'OMFI:PCMA:PeakEnvelopeFormat',        'int32'),
        AVBPropertyDef('points_per_peak_value',       'OMFI:PCMA:PointsPerPeakValue',        'int32'),
        AVBPropertyDef('peak_envelope_block_size',    'OMFI:PCMA:PeakEnvelopeBlockSize',     'int32'),
        AVBPropertyDef('peak_channel_count',          'OMFI:PCMA:PeakChannelCount',          'int32'),
        AVBPropertyDef('peak_frame_count',            'OMFI:PCMA:PeakFrameCount',            'int32'),
        AVBPropertyDef('peak_of_peaks_offset',        'OMFI:PCMA:PeakOfPeaksOffset',         'uint64'),
        AVBPropertyDef('peak_envelope_timestamp',     'OMFI:PCMA:PeakEnvelopeTimestamp',     'int32'),
        AVBPropertyDef('ebu_timestamp',               'OMFI:PCMA:SmpteEbuTimestamp',         'int64'),
        AVBPropertyDef('timecode_framerate',          'OMFI:PCMA:TimecodeFrameRate',        'string'),
    ]
    __slots__ = ()

    def read(self, f):
        super(PCMADescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.channels = ctx.read_u16(f)
        self.quantization_bits = ctx.read_u16(f)
        self.sample_rate = ctx.read_exp10_encoded_float(f)

        self.locked = ctx.read_bool(f)
        self.audio_ref_level = ctx.read_s16(f)
        self.electro_spatial_formulation = ctx.read_s32(f)
        self.dial_norm = ctx.read_u16(f)

        self.coding_format = ctx.read_u32(f)
        self.block_align = ctx.read_u32(f)

        self.sequence_offset = ctx.read_u16(f)
        self.average_bps = ctx.read_u32(f)
        self.has_peak_envelope_data = ctx.read_bool(f)

        self.peak_envelope_version = ctx.read_s32(f)
        self.peak_envelope_format = ctx.read_s32(f)
        self.points_per_peak_value = ctx.read_s32(f)
        self.peak_envelope_block_size = ctx.read_s32(f)
        self.peak_channel_count = ctx.read_s32(f)
        self.peak_frame_count = ctx.read_s32(f)
        self.peak_of_peaks_offset = ctx.read_u64(f)
        self.peak_envelope_timestamp = ctx.read_s32(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 77)
                self.ebu_timestamp = ctx.read_s64(f)
            elif tag == 0x03:
                ctx.read_assert_tag(f, 76)
                # yes this is a string!
                self.timecode_framerate = ctx.read_string(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(PCMADescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u16(f, self.channels)
        ctx.write_u16(f, self.quantization_bits)
        ctx.write_exp10_encoded_float(f, self.sample_rate)

        ctx.write_bool(f, self.locked)
        ctx.write_s16(f, self.audio_ref_level)
        ctx.write_s32(f, self.electro_spatial_formulation)
        ctx.write_u16(f, self.dial_norm)

        ctx.write_u32(f, self.coding_format)
        ctx.write_u32(f, self.block_align)

        ctx.write_u16(f, self.sequence_offset)
        ctx.write_u32(f, self.average_bps)
        ctx.write_bool(f, self.has_peak_envelope_data)

        ctx.write_s32(f, self.peak_envelope_version)
        ctx.write_s32(f, self.peak_envelope_format)
        ctx.write_s32(f, self.points_per_peak_value)
        ctx.write_s32(f, self.peak_envelope_block_size)
        ctx.write_s32(f, self.peak_channel_count)
        ctx.write_s32(f, self.peak_frame_count)
        ctx.write_u64(f, self.peak_of_peaks_offset)
        ctx.write_s32(f, self.peak_envelope_timestamp)

        if hasattr(self, 'ebu_timestamp'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 77)
            ctx.write_s64(f, self.ebu_timestamp)

        if hasattr(self, 'timecode_framerate'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 76)
            ctx.write_string(f, self.timecode_framerate)

        ctx.write_u8(f, 0x03)

@utils.register_class
class DIDDescriptor(MediaFileDescriptor):
    class_id = b'DIDD'
    propertydefs = MediaFileDescriptor.propertydefs + [
        AVBPropertyDef('stored_height',              'OMFI:DIDD:StoredHeight',                            'int32',    1080),
        AVBPropertyDef('stored_width',               'OMFI:DIDD:StoredWidth',                             'int32',    1920),
        AVBPropertyDef('sampled_height',             'OMFI:DIDD:SampledHeight',                           'int32',    1080),
        AVBPropertyDef('sampled_width',              'OMFI:DIDD:SampledWidth',                            'int32',    1920),
        AVBPropertyDef('sampled_x_offset',           'OMFI:DIDD:SampledXOffset',                          'int32',       0),
        AVBPropertyDef('sampled_y_offset',           'OMFI:DIDD:SampledYOffset',                          'int32',       0),
        AVBPropertyDef('display_height',             'OMFI:DIDD:DisplayHeight',                           'int32',    1080),
        AVBPropertyDef('display_width',              'OMFI:DIDD:DisplayWidth',                            'int32',    1920),
        AVBPropertyDef('display_x_offset',           'OMFI:DIDD:DisplayXOffset',                          'int32',       0),
        AVBPropertyDef('display_y_offset',           'OMFI:DIDD:DisplayYOffset',                          'int32',       0),
        AVBPropertyDef('frame_layout',               'OMFI:DIDD:FrameLayout',                             'int16',       0),
        AVBPropertyDef('aspect_ratio',               'OMFI:DIDD:ImageAspectRatio',                        'rational',   [16,9]),
        AVBPropertyDef('line_map',                   'OMFI:DIDD:VideoLineMap',                            'list',       [42,0]),
        AVBPropertyDef('alpha_transparency',         'OMFI:DIDD:AlphaTransparency',                       'int32',       0),
        AVBPropertyDef('uniformness',                'OMFI:DIDD:Uniformness',                             'bool',    False),
        AVBPropertyDef('did_image_size',             'OMFI:DIDD:DIDImageSize',                            'int32', 108511232),
        AVBPropertyDef('next_did_desc',              "OMFI:DIDD:NextDIDDesc",                         'reference',    None),
        AVBPropertyDef('compress_method',            'OMFI:DIDD:DIDCompressMethod',                       'bytes', b'AVHD'),
        AVBPropertyDef('resolution_id',              'OMFI:DIDD:DIDResolutionID',                         'int32',    1237),
        AVBPropertyDef('image_alignment_factor',     'OMFI:DIDD:ImageAlignmentFactor',                    'int32',    8192),
        AVBPropertyDef('frame_index_byte_order',     'OMFI:DIDD:FrameIndexByteOrder',                     'int16'),
        AVBPropertyDef('frame_sample_size',          'OMFI:DIDD:FrameSampleSize',                         'int32'),
        AVBPropertyDef('first_frame_offset',         'OMFI:DIDD:FirstFrameOffset',                        'int32'),
        AVBPropertyDef('client_fill_start',          'OMFI:DIDD:ClientFillStart',                         'int32'),
        AVBPropertyDef('client_fill_end',            'OMFI:DIDD:ClientFillEnd',                           'int32'),
        AVBPropertyDef('offset_to_rle_frame_index',  'OMFI:DIDD:OffsetToRLEFrameIndexes',                 'int32'),
        AVBPropertyDef('frame_start_offset',         'OMFI:DIDD:FrameStartOffset',                        'int32'),
        AVBPropertyDef('valid_box',                  'OMFI:DIDD:Valid',                                   'bounds_box'),
        AVBPropertyDef('essence_box',                'OMFI:DIDD:Essence',                                 'bounds_box'),
        AVBPropertyDef('source_box',                 'OMFI:DIDD:Source',                                  'bounds_box'),
        AVBPropertyDef('framing_box',                'OMFI:DIDD:Framing',                                 'bounds_box'),
        AVBPropertyDef('reformatting_option',        'OMFI:DIDD:ReformattingOption',                      'int32'),
        AVBPropertyDef('transfer_characteristic',    'OMFI:DIDD:TransferCharacteristic',                  'UUID'),
        AVBPropertyDef('color_primaries',            'OMFI:DIDD:ColorPrimaries',                          'UUID'),
        AVBPropertyDef('coding_equations',           'OMFI:DIDD:CodingEquations',                         'UUID'),
        AVBPropertyDef('essence_compression',        'OMFI:DIDD:EssenceCompression',                      'UIDD'),
        AVBPropertyDef('essence_element_size_kind',  'OMFI:DIDD:EssenceElementSizeKind',                 'uint8'),
        AVBPropertyDef('frame_checked_with_mapper',  'OMFI:DIDD:FrameSampleSizeHasBeenCheckedWithMapper', 'bool'),
    ]
    __slots__ = ()

    def read(self, f):
        super(DIDDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        self.stored_height = ctx.read_s32(f)
        self.stored_width  = ctx.read_s32(f)

        self.sampled_height = ctx.read_s32(f)
        self.sampled_width  = ctx.read_s32(f)

        self.sampled_x_offset = ctx.read_s32(f)
        self.sampled_y_offset = ctx.read_s32(f)

        self.display_height = ctx.read_s32(f)
        self.display_width  = ctx.read_s32(f)

        self.display_x_offset = ctx.read_s32(f)
        self.display_y_offset = ctx.read_s32(f)

        self.frame_layout = ctx.read_s16(f)

        numerator = ctx.read_s32(f)
        denominator = ctx.read_s32(f)
        self.aspect_ratio = [numerator, denominator]

        line_map_byte_size = ctx.read_s32(f)
        self.line_map = []
        if line_map_byte_size:
            for i in range(line_map_byte_size // 4):
                v = ctx.read_s32(f)
                self.line_map.append(v)

        self.alpha_transparency = ctx.read_s32(f)
        self.uniformness = ctx.read_bool(f)

        self.did_image_size = ctx.read_s32(f)

        self.next_did_desc = ctx.read_object_ref(self.root, f)

        self.compress_method = ctx.read_fourcc(f)

        self.resolution_id = ctx.read_s32(f)
        self.image_alignment_factor =  ctx.read_s32(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 69)
                self.frame_index_byte_order = ctx.read_s16(f)

            elif tag == 0x02:
                ctx.read_assert_tag(f, 71)
                self.frame_sample_size = ctx.read_s32(f)

            elif tag == 0x03:
                ctx.read_assert_tag(f, 71)
                self.first_frame_offset = ctx.read_s32(f)

            elif tag == 0x04:
                ctx.read_assert_tag(f, 71)
                self.client_fill_start = ctx.read_s32(f)

                ctx.read_assert_tag(f, 71)
                self.client_fill_end = ctx.read_s32(f)

            elif tag == 0x05:
                ctx.read_assert_tag(f, 71)
                self.offset_to_rle_frame_index = ctx.read_s32(f)

            elif tag == 0x06:
                ctx.read_assert_tag(f, 71)
                self.frame_start_offset = ctx.read_s32(f)

            elif tag == 0x08:
                # valid
                self.valid_box = []
                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.valid_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.valid_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.valid_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.valid_box.append([x, y])

                # essence
                self.essence_box = []
                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.essence_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.essence_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.essence_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.essence_box.append([x, y])

                # source
                self.source_box = []
                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.source_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.source_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.source_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.source_box.append([x, y])

            elif tag == 9:
                # print("\n??!", peek_data(f).encode('hex'), '\n')
                self.framing_box = []
                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.framing_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.framing_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.framing_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                x = ctx.read_s32(f)
                ctx.read_assert_tag(f, 71)
                y = ctx.read_s32(f)
                self.framing_box.append([x, y])

                ctx.read_assert_tag(f, 71)
                self.reformatting_option = ctx.read_s32(f)

            elif tag == 10:
                ctx.read_assert_tag(f, 80)
                self.transfer_characteristic = ctx.read_raw_uuid(f)
            elif tag == 11:
                ctx.read_assert_tag(f, 80)
                self.color_primaries =  ctx.read_raw_uuid(f)
                ctx.read_assert_tag(f, 80)
                self.coding_equations = ctx.read_raw_uuid(f)
            elif tag == 12:
                ctx.read_assert_tag(f, 80)
                self.essence_compression = ctx.read_raw_uuid(f)
            elif tag == 14:
                ctx.read_assert_tag(f, 68)
                self.essence_element_size_kind = ctx.read_u8(f)
            elif tag == 15:
                ctx.read_assert_tag(f, 66)
                self.frame_checked_with_mapper = ctx.read_bool(f)

            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] == b'DIDD':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(DIDDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        ctx.write_s32(f, self.stored_height)
        ctx.write_s32(f, self.stored_width)

        ctx.write_s32(f, self.sampled_height)
        ctx.write_s32(f, self.sampled_width)

        ctx.write_s32(f, self.sampled_x_offset)
        ctx.write_s32(f, self.sampled_y_offset)

        ctx.write_s32(f, self.display_height)
        ctx.write_s32(f, self.display_width)

        ctx.write_s32(f, self.display_x_offset)
        ctx.write_s32(f, self.display_y_offset)

        ctx.write_s16(f, self.frame_layout)

        ctx.write_s32(f, self.aspect_ratio[0])
        ctx.write_s32(f, self.aspect_ratio[1])

        ctx.write_s32(f, len(self.line_map) * 4)
        for i in self.line_map:
            ctx.write_s32(f, i)


        ctx.write_s32(f, self.alpha_transparency)
        ctx.write_bool(f, self.uniformness)

        ctx.write_s32(f, self.did_image_size)

        ctx.write_object_ref(self.root, f, self.next_did_desc)

        assert len(self.compress_method) == 4
        ctx.write_fourcc(f, self.compress_method)

        ctx.write_s32(f, self.resolution_id)
        ctx.write_s32(f, self.image_alignment_factor)

        if hasattr(self, 'frame_index_byte_order'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 69)
            ctx.write_s16(f, self.frame_index_byte_order)

        if hasattr(self, 'frame_sample_size'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.frame_sample_size)

        if hasattr(self, 'first_frame_offset'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.first_frame_offset)

        if hasattr(self, 'client_fill_start'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x04)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.client_fill_start)

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.client_fill_end)

        if hasattr(self, 'offset_to_rle_frame_index'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x05)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.offset_to_rle_frame_index)

        if hasattr(self, 'frame_start_offset'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x06)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.frame_start_offset)

        if hasattr(self, 'valid_box') and hasattr(self, 'essence_box') and hasattr(self, 'source_box'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x08)

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[0][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[0][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[1][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[1][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[2][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[2][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[3][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.valid_box[3][1])

            # essence
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[0][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[0][1])


            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[1][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[1][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[2][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[2][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[3][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.essence_box[3][1])

            # source
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[0][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[0][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[1][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[1][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[2][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[2][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[3][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.source_box[3][1])

        if hasattr(self, 'framing_box') and hasattr(self, 'reformatting_option'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 9)

            # print("\n??!", peek_data(f).encode('hex'), '\n')
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[0][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[0][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[1][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[1][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[2][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[2][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[3][0])
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.framing_box[3][1])

            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.reformatting_option)

        if hasattr(self, 'transfer_characteristic'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 10)
            ctx.write_u8(f, 80)
            ctx.write_raw_uuid(f, self.transfer_characteristic)

        if hasattr(self, 'color_primaries') and hasattr(self, 'coding_equations'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 11)
            ctx.write_u8(f, 80)
            ctx.write_raw_uuid(f, self.color_primaries)
            ctx.write_u8(f, 80)
            ctx.write_raw_uuid(f, self.coding_equations)

        if hasattr(self, 'essence_compression'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 12)
            ctx.write_u8(f, 80)
            ctx.write_raw_uuid(f, self.essence_compression)

        if hasattr(self, 'essence_element_size_kind'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 14)
            ctx.write_u8(f, 68)
            ctx.write_u8(f, self.essence_element_size_kind)

        if hasattr(self, 'frame_checked_with_mapper'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 15)
            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.frame_checked_with_mapper)

        if self.class_id[:] == b'DIDD':
            ctx.write_u8(f, 0x03)

@utils.register_class
class CDCIDescriptor(DIDDescriptor):
    class_id = b'CDCI'
    propertydefs = DIDDescriptor.propertydefs + [
        AVBPropertyDef('horizontal_subsampling', 'OMFI:CDCI:HorizontalSubsampling',         'uint32',   2),
        AVBPropertyDef('vertical_subsampling',   'OMFI:CDCI:VerticalSubsampling',           'uint32',   1),
        AVBPropertyDef('component_width',        'OMFI:CDCI:ComponentWidth',                'int32',    8),
        AVBPropertyDef('color_sitting',          'OMFI:CDCI:ColorSiting',                   'int16',    4),
        AVBPropertyDef('black_ref_level',        'OMFI:CDCI:BlackReferenceLevel',           'uint32',  16),
        AVBPropertyDef('white_ref_level',        'OMFI:CDCI:WhiteReferenceLevel',           'uint32', 235),
        AVBPropertyDef('color_range',            'OMFI:CDCI:ColorRange',                    'uint32', 255),
        AVBPropertyDef('frame_index_offset',     'OMFI:JPED:OffsetToFrameIndexes',          'uint64',   0),
        AVBPropertyDef('alpha_sampled_width',    'OMFI:CDCI:AlphaSamledWidth',              'uint32'),
        AVBPropertyDef('ignore_bw',              'OMFI:CDCI:IgnoreBWRefLevelAndColorRange', 'uint32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(CDCIDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x02)

        self.horizontal_subsampling = ctx.read_u32(f)
        self.vertical_subsampling = ctx.read_u32(f)
        self.component_width = ctx.read_s32(f)

        self.color_sitting = ctx.read_s16(f)
        self.black_ref_level = ctx.read_u32(f)
        self.white_ref_level = ctx.read_u32(f)
        self.color_range = ctx.read_u32(f)

        self.frame_index_offset = ctx.read_s64(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 72)
                self.alpha_sampled_width = ctx.read_u32(f)

            elif tag == 0x02:
                ctx.read_assert_tag(f, 72)
                self.ignore_bw = ctx.read_u32(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        if self.class_id[:] == b'CDCI':
            ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(CDCIDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x02)

        ctx.write_u32(f, self.horizontal_subsampling)
        ctx.write_u32(f, self.vertical_subsampling)
        ctx.write_s32(f, self.component_width)

        ctx.write_s16(f, self.color_sitting)
        ctx.write_u32(f, self.black_ref_level)
        ctx.write_u32(f, self.white_ref_level)
        ctx.write_u32(f, self.color_range)

        ctx.write_s64(f, self.frame_index_offset)

        if hasattr(self, 'alpha_sampled_width'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.alpha_sampled_width)

        if hasattr(self, 'ignore_bw'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.ignore_bw)

        if self.class_id[:] == b'CDCI':
            ctx.write_u8(f, 0x03)

@utils.register_class
class MPGIDescriptor(CDCIDescriptor):
    class_id = b'MPGI'
    propertydefs = CDCIDescriptor.propertydefs + [
        AVBPropertyDef('mpeg_version',     "OMFI:MPGI:MPEGVersion",        "uint8"),
        AVBPropertyDef('profile',          "OMFI:MPGI:ProfileAndLevel",    "uint8"),
        AVBPropertyDef('gop_structure',    "OMFI:MPGI:GOPStructure",       "uint8"),
        AVBPropertyDef('stream_type',      "OMFI:MPGI:StreamType",         "uint8"),
        AVBPropertyDef('random_access',    "OMFI:MPGI:RandomAccess",       "bool"),
        AVBPropertyDef('leading_discard',  "OMFI:MPGI:LeadingDiscard",     "bool"),
        AVBPropertyDef('trailing_discard', "OMFI:MPGI:TrailingDiscard",    "bool"),
        AVBPropertyDef('min_gop_length',   "OMFI:MPGI:omMPGIMinGOPLength", "uint16"),
        AVBPropertyDef('max_gop_length',   "OMFI:MPGI:omMPGIMaxGOPLength", "uint16"),
        AVBPropertyDef('sequence_hdr',     "OMFI:MPGI:SequenceHdr",        "bytes"),
    ]
    __slots__ = ()

    def read(self, f):
        super(MPGIDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.mpeg_version = ctx.read_u8(f)
        self.profile = ctx.read_u8(f)
        self.gop_structure = ctx.read_u8(f)
        self.stream_type = ctx.read_u8(f)
        self.random_access = ctx.read_bool(f)
        self.leading_discard = ctx.read_bool(f)
        self.trailing_discard = ctx.read_bool(f)
        self.min_gop_length = ctx.read_u16(f)
        self.max_gop_length = ctx.read_u16(f)
        hdrlen = ctx.read_s32(f)
        assert hdrlen >= 0
        self.sequence_hdr = bytearray(f.read(hdrlen))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(MPGIDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u8(f, self.mpeg_version)
        ctx.write_u8(f, self.profile)
        ctx.write_u8(f, self.gop_structure)
        ctx.write_u8(f, self.stream_type)
        ctx.write_bool(f, self.random_access)
        ctx.write_bool(f, self.leading_discard)
        ctx.write_bool(f, self.trailing_discard)
        ctx.write_u16(f, self.min_gop_length)
        ctx.write_u16(f, self.max_gop_length)
        ctx.write_s32(f, len(self.sequence_hdr))
        f.write(self.sequence_hdr)

        ctx.write_u8(f, 0x03)

@utils.register_class
class JPEGDescriptor(CDCIDescriptor):
    class_id = b'JPED'
    propertydefs = CDCIDescriptor.propertydefs + [
        AVBPropertyDef('jpeg_table_id',           "OMFI:JPED:JPEGTableID",           "int32"),
        AVBPropertyDef('jpeg_frame_index_offset', "OMFI:JPED:OffsetToFrameIndexes",  "uint64"),
        AVBPropertyDef('quantization_tables',     "OMFI:JPED:QuantizationTables",    "bytes"),
        AVBPropertyDef('image_start_align',       "OMFI:JPED:ImageStartAlignment",   "int32"),
    ]
    __slots__ = ()

    def read(self, f):
        super(JPEGDescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.jpeg_table_id = ctx.read_s32(f)
        self.jpeg_frame_index_offset = ctx.read_u64(f)
        table_size = ctx.read_s32(f)
        assert table_size >= 0
        self.quantization_tables = bytearray(f.read(table_size))

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.image_start_align = ctx.read_s32(f)
            else:
                raise ValueError("unknown ext tag 0x%02X %d" % (tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(JPEGDescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.jpeg_table_id)
        ctx.write_u64(f, self.jpeg_frame_index_offset)
        ctx.write_s32(f, len(self.quantization_tables))
        f.write(self.quantization_tables)

        if hasattr(self, 'image_start_align'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.image_start_align)

        ctx.write_u8(f, 0x03)

def encode_pixel_layout(ctx, layout):
    pixel_layout = BytesIO()
    pixel_struct = BytesIO()

    for i in range(len(layout)):
        ctx.write_u8(pixel_layout, layout[i]['Code'])
        ctx.write_u8(pixel_struct, layout[i]['Size'])

    return pixel_layout.getvalue(), pixel_struct.getvalue()

@utils.register_class
class RGBADescriptor(DIDDescriptor):
    class_id = b'RGBA'
    propertydefs = DIDDescriptor.propertydefs + [
        AVBPropertyDef('pixel_layout',       'OMFI:RGBA:PixelLayout',          'list'),
        AVBPropertyDef('palette',            'OMFI:RGBA:Palette',              'list'),
        AVBPropertyDef('frame_index_offset', 'OMFI:RGBA:OffsetToFrameIndexes', 'uint64'),
        AVBPropertyDef('has_comp_min_ref',   'OMFI:RGBA:HasCompMinRef',        'bool'),
        AVBPropertyDef('comp_min_ref',       'OMFI:RGBA:ComponentMinRef',      'uint32'),
        AVBPropertyDef('has_comp_max_ref',   'OMFI:RGBA:HasCompMaxRef',        'bool'),
        AVBPropertyDef('comp_max_ref',       'OMFI:RGBA:ComponentMaxRef',      'uint32'),
        AVBPropertyDef('alpha_min_ref',      'OMFI:RGBA:AlphaMinRef',          'uint32'),
        AVBPropertyDef('alpha_max_ref',      'OMFI:RGBA:AlphaMaxRef',          'uint32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(RGBADescriptor, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        # this seems to be encode the same way as in AAF
        layout_size = ctx.read_u32(f)
        pixel_layout = bytearray(f.read(layout_size))

        struct_size =  ctx.read_u32(f)
        pixel_struct = bytearray(f.read(struct_size))

        assert layout_size == struct_size

        layout = []
        for code, size in zip(pixel_layout, pixel_struct):
            layout.append({'Code': code, "Size" : size})

        self.pixel_layout = layout

        palette_layout_size = ctx.read_u32(f)
        assert palette_layout_size == 0

        palette_struct_size = ctx.read_u32(f)
        assert palette_struct_size == 0

        palette_size = ctx.read_u32(f)
        assert palette_size == 0

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 77)
                self.frame_index_offset = ctx.read_u64(f)

            elif tag == 0x02:
                ctx.read_assert_tag(f, 66)
                self.has_comp_min_ref = ctx.read_bool(f)

                ctx.read_assert_tag(f, 72)
                self.comp_min_ref = ctx.read_u32(f)

                ctx.read_assert_tag(f, 66)
                self.has_comp_max_ref = ctx.read_bool(f)

                ctx.read_assert_tag(f, 72)
                self.comp_max_ref = ctx.read_u32(f)

            elif tag == 0x03:
                ctx.read_assert_tag(f, 72)
                self.alpha_min_ref = ctx.read_u32(f)

                ctx.read_assert_tag(f, 72)
                self.alpha_max_ref = ctx.read_u32(f)

            else:
                raise ValueError("unknown ext tag 0x%02X %d" % (tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(RGBADescriptor, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        pixel_layout, pixel_struct = encode_pixel_layout(ctx, self.pixel_layout)

        # this seems to be encode the same way as in AAF
        ctx.write_u32(f, len(pixel_layout))
        f.write(pixel_layout)

        ctx.write_u32(f, len(pixel_struct))
        f.write(pixel_struct)

        # palette_layout_size
        ctx.write_u32(f, 0)
        # palette_struct_size
        ctx.write_u32(f, 0)
        # palette_size
        ctx.write_u32(f, 0)


        if hasattr(self, 'frame_index_offset'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 77)
            ctx.write_u64(f, self.frame_index_offset)

        if hasattr(self, 'has_comp_min_ref'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.has_comp_min_ref)

            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.comp_min_ref)

            ctx.write_u8(f, 66)
            ctx.write_bool(f, self.has_comp_max_ref)

            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.comp_max_ref)

        if hasattr(self, 'alpha_min_ref'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x03)
            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.alpha_min_ref)

            ctx.write_u8(f, 72)
            ctx.write_u32(f, self.alpha_max_ref)

        ctx.write_u8(f, 0x03)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import sys
from collections import OrderedDict
from . import utils

sentinel = object()

INT_FORMAT = int
if sys.version_info.major < 3:
    INT_FORMAT = (int, long)

class AVBPropertyDef(object):
    __slots__ = ('name', 'long_name', 'type', 'default')
    def __init__(self, name, long_name, data_type, default_value=sentinel):
        self.name = name
        self.long_name = name
        self.type = data_type
        self.default = default_value
    def __repr__(self):
        s = ""
        s += "%s.%s"  % (self.__class__.__module__,
                                self.__class__.__name__)

        s += " " + str(self.name)
        s += " " + str(self.type)

        return '<%s at 0x%x>' % (s, id(self))

class AVBPropertyData(OrderedDict):
    __slots__ = ()
    def deref(self, value):
        if isinstance(value, utils.AVBObjectRef):
            return value.value
        return value

    def __getitem__(self, key):
        return self.deref(super(AVBPropertyData, self).__getitem__(key))

    def items(self):
        for key, value in super(AVBPropertyData, self).items():
            yield key, self.deref(value)

    def get(self, *args, **kwargs):
        return self.deref(super(AVBPropertyData, self).get(*args, **kwargs))

class AVBRefList(list):
    propertydefs = []
    __slots__ = ('root', 'instance_id', '__weakref__')

    def __new__(cls, *args, **kwargs):
        self = super(AVBRefList, cls).__new__(cls)
        self.root = kwargs.get("root", None)
        return self
    # def __init__(self, root):
    #     super(AVBRefList, self).__init__()
    #     self.root = root

    def mark_modified(self):
        if not self.root.reading:
            self.root.add_modified(self)

    def extend(self, x):
        super(AVBRefList, self).extend(x)
        self.mark_modified()

    def append(self, x):
        super(AVBRefList, self).append(x)
        self.mark_modified()

    def insert(self, i, x):
        super(AVBRefList, self).insert(i, x)
        self.mark_modified()

    def remove(self, x):
        super(AVBRefList, self).remove(x)
        self.mark_modified()

    def pop(self, i=-1):
        result = self.deref(super(AVBRefList, self).pop(i))
        self.mark_modified()
        return result

    def clear(self):
        super(AVBRefList, self)
        self.mark_modified()

    def sort(self, key=None, reverse=False):
        super(AVBRefList, self)
        self.mark_modified()

    def reverse(self):
        super(AVBRefList, self)
        self.mark_modified()

    def deref(self, value):
        if isinstance(value, INT_FORMAT):
            return utils.AVBObjectRef(self.root, value).value

        if isinstance(value, utils.AVBObjectRef):
            return value.value
        return value

    def __getitem__(self, index):
        return self.deref(super(AVBRefList, self).__getitem__(index))

    def __setitem__(self, index, value):
        super(AVBRefList, self).__setitem__(index, value)
        self.mark_modified()

    def __delitem__(self, index):
        super(AVBRefList, self).__delitem__(index)
        self.mark_modified()

    def __iter__(self):
        for value in super(AVBRefList, self).__iter__():
            yield self.deref(value)

def walk_references(obj):

    if isinstance(obj, list):
        property_values = obj
    elif isinstance(obj, dict):
        property_values = obj.values()
    elif hasattr(obj, 'property_data'):
        property_values = obj.property_data.values()
    else:
        property_values = []

    for v in property_values:
        if isinstance(v, utils.AVBObjectRef):
            v = v.value
        if v is None:
            continue

        if isinstance(v, list):
            for item in v:
                for sub_v in walk_references(item):
                    yield sub_v

        if hasattr(v, 'class_id') and v.class_id:
            for sub_v in walk_references(v):
                yield sub_v

    if hasattr(obj, 'class_id') and obj.class_id:
        yield obj


class AVBObject(object):
    propertydefs = []
    propertydefs_dict = None
    class_id = None
    __slots__ = ('root', 'property_data', 'instance_id', '__weakref__')

    def __new__(cls, *args, **kwargs):
        self = super(AVBObject, cls).__new__(cls)
        self.root = kwargs.get("root", None)
        self.property_data = AVBPropertyData()
        return self

    def __init__(self, *args, **kwargs):
        for pdef in self.propertydefs:
            if pdef.default is sentinel:
                continue
            self.__setattr__(pdef.name, pdef.default)

    def mark_modified(self):
        if not self.root.reading and hasattr(self, 'instance_id'):
            self.root.add_modified(self)

    def get_property_def(self, name):
        for item in self.propertydefs:
            if item.name == name:
                return item

    def get(self, key, default):
        for property_key, value in self.property_data.items():
            if property_key == key:
                return value
        return default

    def __setattr__(self, name, value):
        pdef = self.propertydefs_dict.get(name, None)
        if pdef:
            self.property_data[name] = value
            self.mark_modified()
            return

        super(AVBObject, self).__setattr__(name, value)

    def __getattr__(self, name):
        v = self.property_data.get(name, sentinel)
        if v is not sentinel:
            return v

        raise AttributeError("'%s' has no attribute '%s'" % (self.__class__.__name__, name))

    def read(self, f):
        pass

    def write(self, f):
        pass

    @property
    def media_kind(self):
        return None

    def __repr__(self):
        s = "%s.%s"  % (self.__class__.__module__,
                                self.__class__.__name__)

        if self.class_id:
            s = str(self.class_id) + " " + s

        if 'name' in self.property_data and self.name:
            s += " " + self.name

        if self.class_id == b'CMPO' and self.mob_type:
            s += " " + self.mob_type

        if 'effect_id' in self.property_data and self.effect_id:

            s += " " + self.effect_id

        if self.class_id and self.media_kind:
            s += " "  + str(self.media_kind)

        if 'length' in self.property_data:
            s += " len: " + str(self.length)

        if 'track_id' in self.property_data:
            s += ' track_id: ' + str(self.track_id)

        if 'start_time' in self.property_data:
            s += ' start_time: ' + str(self.start_time)

        if 'mob_id' in self.property_data:
            s += " " + str(self.mob_id)

        if 'uuid' in self.property_data:
            s += " " + str(self.uuid)

        if 'value' in self.property_data:
            s += " " + str(self.value)

        return '<%s at 0x%x>' % (s, id(self))



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from . import utils
from . import core
from .core import AVBPropertyDef, AVBPropertyData, AVBRefList
from . utils import peek_data

INT_ATTR  = 1
STR_ATTR  = 2
OBJ_ATTR  = 3
BOB_ATTR  = 4

if bytes is not str:
    unicode = str

@utils.register_class
class Attributes(AVBPropertyData):
    class_id = b'ATTR'
    propertydefs = []
    __slots__ = ('root', 'instance_id', )

    def __new__(cls, *args, **kwargs):
        self = super(Attributes, cls).__new__(cls)
        self.root = kwargs.get('root', None)
        return self

    def mark_modified(self):
        if not self.root.reading:
            self.root.add_modified(self)

    def __setitem__(self, key, value):
        super(Attributes, self).__setitem__(key, value)
        self.mark_modified()

    def __delitem__(self, key):
        super(Attributes, self).__delitem__(key)
        self.mark_modified()

    def clear(self):
        super(Attributes, self).clear()
        self.mark_modified()

    def pop(self, *args, **kwargs):
        result = super(Attributes, self).pop(*args, **kwargs)
        self.mark_modified()
        return result

    def read(self, f):
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        count = ctx.read_u32(f)


        for i in range(count):
            attr_type = ctx.read_u32(f)
            attr_name = ctx.read_string(f)

            if attr_type == INT_ATTR:
                self[attr_name] = ctx.read_s32(f)
            elif attr_type == STR_ATTR:
                self[attr_name] = ctx.read_string(f)
            elif attr_type == OBJ_ATTR:
                self[attr_name] = ctx.read_object_ref(self.root, f)
            elif attr_type == BOB_ATTR:
                size = ctx.read_u32(f)
                self[attr_name] = bytearray(f.read(size))
            else:
                raise Exception("Unkown attr name: %s type: %d" % ( attr_name, attr_type))

        # print("read", result)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u32(f, len(self))
        for key, value in self.items():
            assert isinstance(key, unicode)

            if isinstance(value, int):
                attr_type = INT_ATTR
            elif isinstance(value, unicode):
                attr_type = STR_ATTR
            elif isinstance(value, bytearray):
                attr_type = BOB_ATTR
            elif isinstance(value, bytes):
                raise ValueError("%s: bytes value type too ambiguous, use bytearray or unicode str" % key)
            else:
                # assume its AVBObject for now and hope for the best :p
                attr_type = OBJ_ATTR

            ctx.write_u32(f, attr_type)
            ctx.write_string(f, key)

            if attr_type == INT_ATTR:
                ctx.write_s32(f, value)
            elif attr_type == STR_ATTR:
                ctx.write_string(f, value)
            elif attr_type == OBJ_ATTR:
                ctx.write_object_ref(self.root, f, value)
            elif attr_type == BOB_ATTR:
                ctx.write_u32(f, len(value))
                f.write(value)

        ctx.write_u8(f, 0x03)

@utils.register_class
class ParameterList(AVBRefList):
    class_id = b'PRLS'
    __slots__ = ()
    def read(self, f):
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        count = ctx.read_s32(f)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.append(ref)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, len(self))
        for obj in self:
            ctx.write_object_ref(self.root, f, obj)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TimeCrumbList(AVBRefList):
    class_id = b'TMCS'
    __slots__ = ()

    def read(self, f):
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        count = ctx.read_s16(f)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            self.append(ref)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, len(self))
        for obj in self:
            ctx.write_object_ref(self.root, f, obj)

        ctx.write_u8(f, 0x03)



import sys
import os
from setuptools import setup
import setuptools.command.build_py
from distutils.extension import Extension

PROJECT_METADATA = {
    "version": "0.1.0.dev7",
    "author": 'Mark Reid',
    "author_email": 'mindmark@gmail.com',
    "license": 'MIT',
}

METADATA_TEMPLATE = """
__version__ = "{version}"
__author__ = "{author}"
__author_email__ = "{author_email}"
__license__ = "{license}"
"""

sourcefiles = [
"avb/_ext.pyx",
]

extensions =[]
try:
    from Cython.Build import cythonize
    if int(os.environ.get("PYAVB_BUILD_EXT", '1')):
        extensions = cythonize([Extension("avb._ext",
                                sourcefiles,
                                language="c++")])
except ImportError as e:
    print('unable to build optional cython extension')


class AddMetadata(setuptools.command.build_py.build_py):
    """Stamps PROJECT_METADATA into __init__ files."""

    def run(self):
        setuptools.command.build_py.build_py.run(self)

        if self.dry_run:
            return

        target_file = os.path.join(self.build_lib, 'avb', "__init__.py")
        source_file = os.path.join(os.path.dirname(__file__), 'avb', "__init__.py")

        # get the base data from the original file
        with open(source_file, 'r') as fi:
            src_data = fi.read()

        # write that + the suffix to the target file
        with open(target_file, 'w') as fo:
            fo.write(src_data)
            fo.write(METADATA_TEMPLATE.format(**PROJECT_METADATA))

setup(
    name='pyavb',
    description='A python module for the reading and writing Avid Bin Files files.',
    url='https://github.com/markreidvfx/pyavb',

    project_urls={
        'Source':
            'https://github.com/markreidvfx/pyavb',
        'Documentation':
            'http://pyavb.readthedocs.io',
        'Issues':
            'https://github.com/markreidvfx/pyavb/issues',
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Display',
        'Topic :: Multimedia :: Video :: Non-Linear Editor',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],

    keywords='film tv editing editorial edit non-linear edl time',

    platforms='any',

    packages=[
        'avb',
    ],

    cmdclass={'build_py': AddMetadata},
    ext_modules = extensions,
    extras_require= {'cython' : ['cython']},

    **PROJECT_METADATA
)



# cython: language_level=3, distutils: language = c++, boundscheck=False, profile=False

from libcpp.vector cimport vector
cimport cython

IF UNAME_SYSNAME == "Windows":
    cdef extern from *:
        ctypedef unsigned char  uint8_t
        ctypedef   signed char  int8_t

        ctypedef unsigned short uint16_t
        ctypedef   signed short int16_t

        ctypedef unsigned int   uint32_t
        ctypedef   signed int   int32_t

        ctypedef unsigned long long int uint64_t
        ctypedef   signed long long int int64_t
ELSE:
    from libc.stdint cimport (uint8_t, int16_t, uint32_t, int32_t, uint64_t, int64_t)

from datetime import datetime
import uuid

from .utils import AVBObjectRef
from . import utils
from .mobid import MobID
from . import core

cdef extern from "" namespace "Properties":
    struct ChildData:
        const char *name
        vector[Properties] data

cdef extern from "_ext_core.cpp" nogil:
    cdef enum StringType:
        MACROMAN,
        UTF8

    cdef enum PropertyType:
        TRKG,
        TRACK,
        PARAM,

    cdef enum AttrType:
        INT_ATTR,
        STR_ATTR,
        OBJ_ATTR,
        BOB_ATTR,

    cdef enum ControlPointType:
        ParamControlPointType,

    cdef enum ControlPointValueType:
        CP_TYPE_INT,
        CP_TYPE_DOUBLE,
        CP_TYPE_REFERENCE

    cdef union IntDataValue:
        uint64_t u64
        int64_t  s64

    cdef struct IntData:
        const char *name
        bint is_signed;
        IntDataValue data

    cdef struct IntArrayData:
        const char *name
        vector[int64_t] data

    cdef struct DoubleData:
        const char *name
        double data

    cdef struct BoolData:
        const char *name
        bint data

    cdef struct StringData:
        const char *name;
        StringType type
        vector[uint8_t] data

    cdef struct BytesData:
        const char *name;
        vector[uint8_t] data

    cdef struct RefListData:
        const char *name
        vector[uint32_t] data

    cdef struct PerPoint:
        int16_t code;
        ControlPointValueType type
        uint32_t value;
        double double_value;

    cdef struct ControlPoint:
        int32_t offset_num;
        int32_t offset_den;
        int32_t timescale;
        uint32_t value;
        double double_value;
        vector[PerPoint] pp;

    cdef struct ControlPointData:
        const char *name
        ControlPointType type
        ControlPointValueType value_type
        vector[ControlPoint] data

    cdef struct AttrData:
        vector[uint8_t] name
        AttrType type
        uint32_t value
        vector[uint8_t] data

    cdef struct Buffer:
        const uint8_t *root
        const uint8_t *ptr
        const uint8_t *end
        const char *error_message

    cdef struct Properties:
        PropertyType type
        vector[IntData]   refs
        vector[IntData]   ints
        vector[BoolData]   bools
        vector[IntData]   dates
        vector[DoubleData] doubles
        vector[BytesData]  uuids
        vector[StringData] strings
        vector[BytesData]  mob_ids
        vector[RefListData] reflists
        vector[ChildData]  children;
        vector[ControlPointData] control_points
        vector[IntArrayData] arrays
        vector[BytesData] bytearrays

    cdef int read_attributes(Buffer *f, vector[AttrData] &d) except+
    cdef int read_comp(Buffer *f, Properties *p) except+
    cdef int read_sequence(Buffer *f, Properties *p) except+
    cdef int read_sourceclip(Buffer *f, Properties *p) except+
    cdef int read_paramclip(Buffer *f, Properties *p) except+
    cdef int read_paramitem(Buffer *f, Properties *p) except+
    cdef int read_trackref(Buffer *f, Properties *p) except+
    cdef int read_filler(Buffer *f, Properties *p) except+
    cdef int read_trackeffect(Buffer *f, Properties *p) except+
    cdef int read_selector(Buffer *f, Properties *p) except+
    cdef int read_composition(Buffer *f, Properties *p) except+
    cdef int read_media_descriptor(Buffer *f, Properties *p) except+
    cdef int read_did_descriptor(Buffer *f, Properties *p) except+
    cdef int read_cdci_descriptor(Buffer *f, Properties *p) except+
    cdef int read_effectparamlist(Buffer *f, Properties *p) except+

cdef class AVBPropertyData(dict):

    def deref(self, value):
        if isinstance(value, utils.AVBObjectRef):
            return value.value
        return value

    def __getitem__(self, key):
        return self.deref(super(AVBPropertyData, self).__getitem__(key))

    def items(self):
        for key, value in super(AVBPropertyData, self).items():
            yield key, self.deref(value)

    def get(self, *args, **kwargs):
        return self.deref(super(AVBPropertyData, self).get(*args, **kwargs))

cdef void refs2dict(object root, dict d, Properties* p):
    cdef IntData item
    for item in p.refs:
        d[item.name.decode('utf-8')] = AVBObjectRef(root, item.data.u64)

cdef void reflist2dict(object root, dict d, Properties* p):
    cdef RefListData item
    cdef uint32_t i

    for item in p.reflists:
        reflist = core.AVBRefList.__new__(core.AVBRefList, root=root)
        reflist.extend(item.data)
        d[item.name.decode('utf-8')] = reflist

cdef void controlpoints2dict(object root, dict d, Properties* p):

    cdef ControlPointData item
    cdef ControlPoint cp
    cdef PerPoint pp
    cdef bytes name
    cdef list control_point_list
    cdef list pp_list

    cdef object obj_class
    cdef object pp_obj_class

    cdef dict cpdata
    cdef dict ppdata

    for item in p.control_points:
        if item.type == ParamControlPointType:
            obj_class = utils.AVBClassName_dict['ParamControlPoint']
            pp_obj_class = utils.AVBClassName_dict['ParamPerPoint']
        else:
            obj_class = utils.AVBClassName_dict['ControlPoint']
            pp_obj_class = utils.AVBClassName_dict['PerPoint']

        control_point_list = []
        for cp in item.data:

            cpdata = AVBPropertyData()
            cpdata['offset'] = (cp.offset_num, cp.offset_den)
            cpdata['timescale'] = cp.timescale

            if item.value_type == CP_TYPE_INT:
                cpdata['value'] = cp.value
            elif item.value_type == CP_TYPE_DOUBLE:
                cpdata['value'] =cp.double_value
            elif item.value_type == CP_TYPE_REFERENCE:
                cpdata['value'] = utils.AVBObjectRef(root, cp.value)

            pp_list = []
            for pp in cp.pp:
                ppdata = AVBPropertyData()
                ppdata['code'] = pp.code
                ppdata['type'] = pp.type
                if pp.type == CP_TYPE_DOUBLE:
                    ppdata['value'] = pp.double_value
                elif pp.type == CP_TYPE_INT:
                    ppdata['value'] = pp.value

                py_pp = pp_obj_class.__new__(pp_obj_class, root=root)
                py_pp.property_data = ppdata
                pp_list.append(py_pp)

            cpdata['pp'] = pp_list
            py_cp = obj_class.__new__(obj_class, root=root)
            py_cp.property_data = cpdata
            control_point_list.append(py_cp)

        d[item.name.decode("utf-8")] = control_point_list


cdef void ints2dict(dict d, Properties* p):
    cdef IntData item
    cdef int64_t signed_value
    for item in p.ints:
        if item.is_signed:
            d[item.name.decode('utf-8')] = item.data.s64
        else:
            d[item.name.decode('utf-8')] = item.data.u64

cdef void dates2dict(dict d, Properties* p):
    cdef IntData item
    for item in p.dates:
        d[item.name.decode('utf-8')] = datetime.fromtimestamp(item.data.u64)

cdef void doubles2dict(dict d, Properties *p):
    cdef DoubleData item
    for item in p.doubles:
        name = <bytes> item.name
        d[item.name.decode('utf-8')] = item.data

cdef void bools2dict(dict d,  Properties *p):
    cdef BoolData item
    for item in p.bools:
        d[item.name.decode('utf-8')] = item.data

cdef void uuid2dict(dict d, Properties *p):
    cdef bytes data

    cdef BytesData item
    cdef uint8_t *ptr

    for item in p.uuids:
        ptr = &item.data[0]
        data = <bytes> ptr[:16]

        d[item.name.decode('utf-8')] = uuid.UUID(bytes_le=data)

cdef void mob_id2dict(dict d, Properties *p):
    cdef bytes data
    cdef BytesData item
    cdef uint8_t *ptr

    for item in p.mob_ids:
        ptr = &item.data[0]
        data = <bytes> ptr[:32]

        d[item.name.decode('utf-8')] = MobID(bytes_le=data)

cdef void strings2dict(dict d, Properties *p):
    cdef const char * ptr

    cdef StringData item
    cdef size_t data_size;

    for item in p.strings:
        data_size = item.data.size()

        if data_size > 0:
            ptr = <const char *>&item.data[0]
            d[item.name.decode('utf-8')] = ptr[:data_size].decode("macroman")
        else:
            d[item.name.decode('utf-8')] = u""

cdef void children2dict(object root, dict d, Properties *p):
    cdef ChildData item
    cdef Properties child_properties

    cdef list plist
    cdef dict pdata

    for item in p.children:
        plist = []

        for child_properties in item.data:
            pdata = process_poperties(root, &child_properties)

            if child_properties.type == TRACK:
                obj_class = utils.AVBClassName_dict['Track']
            elif child_properties.type == PARAM:
                obj_class = utils.AVBClassName_dict['EffectParam']

            object_instance = obj_class.__new__(obj_class, root=root)
            object_instance.property_data = pdata
            plist.append(object_instance)

        d[item.name.decode('utf-8')] = plist

cdef void int_array2dict(dict d, Properties *p):
    cdef IntArrayData item
    for item in p.arrays:
        name = item.name.decode('utf-8')
        if name in ('valid_box', 'essence_box', 'source_box', 'framing_box'):
            d[name] = [ [item.data[0], item.data[1]],
                        [item.data[2], item.data[3]],
                        [item.data[4], item.data[5]],
                        [item.data[6], item.data[7]],]
        else:
            d[name] = item.data

cdef void bytearray2dict(dict d, Properties *p):
    cdef BytesData item
    cdef uint8_t *ptr
    cdef size_t size
    for item in p.bytearrays:
        size = item.data.size()
        if size:
            ptr = &item.data[0]
            d[item.name.decode('utf-8')] = <bytearray> ptr[:size]
        else:
            d[item.name.decode('utf-8')] = bytearray()

cdef dict process_poperties(object root, Properties *p):
    cdef dict result = AVBPropertyData()

    if p.refs.size():
        refs2dict(root, result, p)
    if p.reflists.size():
        reflist2dict(root, result, p)
    if p.ints.size():
        ints2dict(result, p)
    if p.doubles.size():
        doubles2dict(result, p)
    if p.strings.size():
        strings2dict(result, p)
    if p.bools.size():
        bools2dict(result, p)
    if p.uuids.size():
        uuid2dict(result, p)
    if p.mob_ids.size():
        mob_id2dict(result, p)
    if p.dates.size():
        dates2dict(result, p)
    if p.arrays.size():
        int_array2dict(result, p)
    if p.bytearrays.size():
        bytearray2dict(result, p)
    if p.control_points.size():
        controlpoints2dict(root, result, p)
    if p.children.size():
        children2dict(root, result, p)

    return result


def read_attr_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef AttrData item
    cdef vector[AttrData] d

    cdef object value

    cdef const char * ptr
    cdef size_t data_size;
    cdef int ret = 0

    with nogil:
        ret = read_attributes(&buf, d)

    if ret < 0:
        raise ValueError("Error reading Attributes: %s" % buf.error_message.decode("utf-8"))

    for item in d:

        value = None

        if item.type == INT_ATTR:
            value = <int32_t>item.value

        elif item.type == STR_ATTR:
            data_size = item.data.size()
            ptr =  <const char *>&item.data[0]
            value = ptr[:data_size].decode('macroman')

        elif item.type == OBJ_ATTR:
            value = utils.AVBObjectRef(root, item.value)

        elif item.type == BOB_ATTR:
            data_size = item.data.size()
            ptr =  <const char *>&item.data[0]
            value = <bytearray> ptr[:data_size]

        data_size = item.name.size()
        ptr =  <const char *>&item.name[0]
        object_instance[ptr[:data_size].decode('macroman')] = value

# cdef print_property_sizes(Properties *p):
#     print("refs",           p.refs.size())
#     print("ints",           p.ints.size())
#     print("bools",          p.bools.size())
#     print("dates",          p.dates.size())
#     print("doubles",        p.doubles.size())
#     print("uuids",          p.uuids.size())
#     print("strings",        p.strings.size())
#     print("mob_ids",        p.mob_ids.size())
#     print("reflists",       p.reflists.size())
#     print("children",       p.children.size())
#     print("control_points", p.control_points.size())

def read_sequence_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_sequence(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading SEQU: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_sourceclip_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_sourceclip(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading SCLP: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_paramclip_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        p.refs.reserve(6)
        p.ints.reserve(6)
        ret = read_paramclip(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading PRCL: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_paramitem_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_paramitem(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading PRIT: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_trackref_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_trackref(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading TRKR: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_filler_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_filler(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading FILL: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def reads_selector_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_selector(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading SLCT: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_composition_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        p.ints.reserve(6)
        p.refs.reserve(7)
        p.strings.reserve(2)
        ret = read_composition(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading CMPO: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_media_descriptor_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_media_descriptor(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading MDES: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_did_descriptor_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_did_descriptor(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading DIDD: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result


def read_cdci_descriptor_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_cdci_descriptor(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading CDCI: %s" % buf.error_message.decode("utf-8"))

    # print_property_sizes(&p)
    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result

def read_trackeffect_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        p.refs.reserve(8)
        p.ints.reserve(12)
        ret = read_trackeffect(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading TKFX: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    # print_property_sizes(&p)
    object_instance.property_data = result

def read_effectparamlist_data(root, object_instance, const unsigned char[:] data):
    cdef Buffer buf
    buf.root = &data[0]
    buf.ptr =  &data[0]
    buf.end = &data[-1]
    buf.error_message = ""

    cdef Properties p
    cdef int ret
    with nogil:
        ret = read_effectparamlist(&buf, &p)

    if ret < 0:
        raise ValueError("Error reading FXPS: %s" % buf.error_message.decode("utf-8"))

    cdef dict result = process_poperties(root, &p)

    object_instance.property_data = result


READERS = {
b'CMPO': read_composition_data,
b'TKFX': read_trackeffect_data,
b'MDES': read_media_descriptor_data,
b'DIDD': read_did_descriptor_data,
b'CDCI': read_cdci_descriptor_data,
b'SLCT': reads_selector_data,
b"SEQU": read_sequence_data,
b'FILL': read_filler_data,
b'SCLP': read_sourceclip_data,
b'PRCL': read_paramclip_data,
b'PRIT': read_paramitem_data,
b'TRKR': read_trackref_data,
b'FXPS': read_effectparamlist_data,
b'ATTR': read_attr_data,
}



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

from . import core
from .core import AVBPropertyDef, AVBRefList
from . import utils
from . import mobid
from . utils import peek_data

class Component(core.AVBObject):
    class_id = b'COMP'
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('left_bob',      '__OMFI:CPNT:LeftBob',    'reference', None),
        AVBPropertyDef('right_bob',     '__OMFI:CPNT:RightBob',   'reference', None),
        AVBPropertyDef('media_kind_id', 'OMFI:CPNT:TrackKind',    'int16',        0),
        AVBPropertyDef('edit_rate',     'EdRate',                 'fexp10',      25),
        AVBPropertyDef('name',          'OMFI:CPNT:Name',         'string',    None),
        AVBPropertyDef('effect_id',     'OMFI:CPNT:EffectID',     'string',    None),
        AVBPropertyDef('attributes',    'OMFI:CPNT:Attributes',   'reference'),
        AVBPropertyDef('session_attrs', 'OMFI:CPNT:SessionAttrs', 'reference', None),
        AVBPropertyDef('precomputed',   'OMFI:CPNT:Precomputed',  'reference', None),
        AVBPropertyDef('param_list',    'OMFI:CPNT:ParamList',    'reference'),
    ]
    __slots__ = ()

    def __init__(self, edit_rate=25, media_kind=None):
        super(Component, self).__init__(self)
        self.attributes = self.root.create.Attributes()
        self.media_kind = media_kind
        self.edit_rate = edit_rate

    def read(self, f):
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        # bob == bytes of binary or bag of bits?
        self.left_bob  = ctx.read_object_ref(self.root, f)
        self.right_bob = ctx.read_object_ref(self.root, f)

        self.media_kind_id = ctx.read_s16(f)
        self.edit_rate = ctx.read_exp10_encoded_float(f)
        self.name = ctx.read_string(f) or None
        self.effect_id = ctx.read_string(f) or None

        self.attributes = ctx.read_object_ref(self.root, f)
        self.session_attrs = ctx.read_object_ref(self.root, f)

        self.precomputed = ctx.read_object_ref(self.root, f)

        for tag in ctx.iter_ext(f):

            if tag == 0x01:
                ctx.read_assert_tag(f, 72)
                self.param_list = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

    def write(self, f):
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)
        ctx.write_object_ref(self.root, f, self.left_bob)
        ctx.write_object_ref(self.root, f, self.right_bob)
        ctx.write_s16(f, self.media_kind_id)

        ctx.write_exp10_encoded_float(f, self.edit_rate)

        if self.name:
            ctx.write_string(f, self.name)
        else:
            ctx.write_u16(f, 0xFFFF)

        if self.effect_id:
            ctx.write_string(f, self.effect_id)
        else:
            ctx.write_u16(f, 0xFFFF)

        ctx.write_object_ref(self.root, f, self.attributes)
        ctx.write_object_ref(self.root, f, self.session_attrs)
        ctx.write_object_ref(self.root, f, self.precomputed)

        if hasattr(self, 'param_list'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 72)
            ctx.write_object_ref(self.root, f, self.param_list)

    @property
    def media_kind(self):
        if self.media_kind_id == 0:
            return None
        elif self.media_kind_id == 1:
            return "picture"
        elif self.media_kind_id == 2:
            return "sound"
        elif self.media_kind_id == 3:
            return "timecode"
        elif self.media_kind_id == 4:
            return "edgecode"
        elif self.media_kind_id == 5:
            return "attribute"
        elif self.media_kind_id == 6:
            return 'effectdata'
        elif self.media_kind_id == 7:
            return 'DescriptiveMetadata'
        else:
            return "unknown%d" % self.media_kind_id

    @media_kind.setter
    def media_kind(self, value):
        if value == None:
            self.media_kind_id = 0
        elif value == "picture":
            self.media_kind_id = 1
        elif value == "sound":
            self.media_kind_id = 2
        elif value =="timecode":
            self.media_kind_id = 3
        elif value == "edgecode":
            self.media_kind_id = 4
        elif value == "attribute":
            self.media_kind_id = 5
        elif value == 'effectdata':
            self.media_kind_id = 6
        elif value == 'DescriptiveMetadata':
            self.media_kind_id = 7
        else:
            raise ValueError('unknown media kind: %s' % str(value))

@utils.register_class
class Sequence(Component):
    class_id = b"SEQU"
    propertydefs_dict = {}
    propertydefs = Component.propertydefs + [
        AVBPropertyDef('components', 'OMFI:SEQU:Sequence', 'ref_list'),
    ]
    __slots__ = ()

    def __init__(self, edit_rate=25, media_kind=None):
        super(Sequence, self).__init__(edit_rate=edit_rate, media_kind=media_kind)
        self.components = []

    def read(self, f):
        super(Sequence, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        count = ctx.read_u32(f)
        self.components = AVBRefList.__new__(AVBRefList, root=self.root)
        for i in range(count):
            ref = ctx.read_object_ref(self.root, f)
            # print ref
            self.components.append(ref)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Sequence, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_u32(f, len(self.components))
        for c in self.components:
            ctx.write_object_ref(self.root, f, c)

        ctx.write_u8(f, 0x03)

    @property
    def length(self):
        length = 0
        for component in self.components:
            if component.class_id == b'TNFX':
                length -= component.length
            else:
                length += component.length
        return length

    def nearest_component_at_time(self, edit_unit):
        """ returns the nearest component to edit_unit and its start position"""
        index, position = self.nearest_index_at_time(edit_unit)
        return self.components[index], position

    def nearest_index_at_time(self, edit_unit):
        """returns the index of the nearest component to edit_unit and its start position"""
        last_component = None
        last_index = None
        last_pos = 0

        if edit_unit <= 0:
            return 0

        # this needs to go past target index to handle Transitions
        for index, position, component in self.positions():

            if component.class_id == b'TNFX':
                if position <= edit_unit < position + component.length:
                    return index, position

            # gone past return previous
            if last_component and position >= edit_unit:
                return last_index, last_pos

            last_component = component
            last_index = index
            last_pos = position

        return last_index, last_pos

    def positions(self):
        length = 0
        for index, component in enumerate(self.components):

            if component.class_id == b'TNFX':
                length -= component.length
                yield (index, length, component)
            else:
                yield (index, length, component)
                length += component.length


class Clip(Component):
    class_id = b'CLIP'
    propertydefs_dict = {}
    propertydefs = Component.propertydefs + [
        AVBPropertyDef('length', 'OMFI:CLIP:Length', 'int32', 0),
    ]
    __slots__ = ()

    def read(self, f):
        super(Clip, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)
        self.length = ctx.read_u32(f)

    def write(self, f):
        super(Clip, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)
        ctx.write_u32(f, self.length)

@utils.register_class
class SourceClip(Clip):
    class_id = b'SCLP'
    propertydefs_dict = {}
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('track_id',   'OMFI:SCLP:SourceTrack',     'int16', 0),
        AVBPropertyDef('start_time', 'OMFI:SCLP:SourcePosition',  'int32', 0),
        AVBPropertyDef('mob_id',     'MobID',                     'MobID'),
    ]
    __slots__ = ()

    def __init__(self, edit_rate=25, media_kind=None):
        super(SourceClip, self).__init__(edit_rate=edit_rate, media_kind=media_kind)
        self.mob_id = mobid.MobID()

    def read(self, f):
        super(SourceClip, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        mob_id_hi = ctx.read_u32(f)
        mob_id_lo = ctx.read_u32(f)

        self.track_id = ctx.read_s16(f)
        self.start_time = ctx.read_s32(f)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                self.mob_id = ctx.read_mob_id(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(SourceClip, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        lo = self.mob_id.material.time_low
        hi = self.mob_id.material.time_mid + (self.mob_id.material.time_hi_version << 16)

        ctx.write_u32(f, lo)
        ctx.write_u32(f, hi)

        ctx.write_s16(f, self.track_id)
        ctx.write_s32(f, self.start_time)

        if hasattr(self, 'mob_id'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_mob_id(f, self.mob_id)

        ctx.write_u8(f, 0x03)

    @property
    def mob(self):
        if hasattr(self, 'mob_id'):
            mob_id = self.mob_id
            if mob_id:
                return self.root.content.mob_dict.get(self.mob_id, None)

    @property
    def track(self):
        mob = self.mob
        if mob:
            for track in mob.tracks:
                if track.index == self.track_id and track.component and self.media_kind == track.component.media_kind:
                    return track

@utils.register_class
class Timecode(Clip):
    class_id = b'TCCP'
    propertydefs_dict = {}
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('flags', 'OMFI:TCCP:Flags',   'int32',  0),
        AVBPropertyDef('fps',   'OMFI:TCCP:FPS',     'int32', 25),
        AVBPropertyDef('start', 'OMFI:TCCP:StartTC', 'int32',  0),
    ]
    __slots__ = ()

    def read(self, f):
        super(Timecode, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        # drop ??
        self.flags = ctx.read_u32(f)
        self.fps = ctx.read_u16(f)

        # unused
        f.read(6)

        self.start = ctx.read_u32(f)
        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Timecode, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u32(f, self.flags)
        ctx.write_u16(f, self.fps)
        f.write(bytearray(6))
        ctx.write_u32(f, self.start)

        ctx.write_u8(f, 0x03)

@utils.register_class
class Edgecode(Clip):
    class_id = b'ECCP'
    propertydefs_dict = {}
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('header',      'OMFI:ECCP:Header',      'bytes'),
        AVBPropertyDef('film_kind',   'OMFI:ECCP:FilmKind',   'uint8'),
        AVBPropertyDef('code_format', 'OMFI:ECCP:CodeFormat', 'uint8'),
        AVBPropertyDef('base_perf',   'OMFI:ECCP:BasePerf',   'uint16'),
        AVBPropertyDef('start_ec',    'OMFI:ECCP:StartEC',    'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(Edgecode, self).read(f)
        ctx = self.root.ictx

        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.header = bytearray(f.read(8))
        self.film_kind = ctx.read_u8(f)
        self.code_format =  ctx.read_u8(f)
        self.base_perf = ctx.read_u16(f)
        unused_a  = ctx.read_u32(f)
        self.start_ec = ctx.read_s32(f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Edgecode, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        assert len(self.header) == 8
        f.write(self.header)
        ctx.write_u8(f, self.film_kind)
        ctx.write_u8(f, self.code_format)
        ctx.write_u16(f, self.base_perf)
        #unused
        ctx.write_u32(f, 0)
        ctx.write_s32(f, self.start_ec)

        ctx.write_u8(f, 0x03)

@utils.register_class
class TrackRef(Clip):
    class_id = b'TRKR'
    propertydefs_dict = {}
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('relative_scope', 'OMFI:TRKR:RelativeScope', 'int16',  0),
        AVBPropertyDef('relative_track', 'OMFI:TRKR:RelativeTrack', 'int16', -1),
    ]
    __slots__ = ()

    def read(self, f):
        super(TrackRef, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.relative_scope = ctx.read_s16(f)
        self.relative_track = ctx.read_s16(f)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(TrackRef, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s16(f, self.relative_scope)
        ctx.write_s16(f, self.relative_track)

        ctx.write_u8(f, 0x03)


CP_TYPE_INT = 1
CP_TYPE_DOUBLE = 2
CP_TYPE_REFERENCE = 4

@utils.register_helper_class
class ParamControlPoint(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('offset',    'OMFI:PRCL:Offset',     'rational'),
        AVBPropertyDef('timescale', 'OMFI:PRCL:TimeScale',  'int32'),
        AVBPropertyDef('value',     'OMFI:PRCL:Value',      'number'), # int or double
        AVBPropertyDef('pp',        'OMFI:PRCL:PP',         'list'),
    ]
    __slots__ = ()

# not sure hwat PP's stands for
@utils.register_helper_class
class ParamPerPoint(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('code',  'OMFI:PRCL:PPCode',  'int16'),
        AVBPropertyDef('type',  'OMFI:PRCL:PPType',  'int16'),
        AVBPropertyDef('value', 'OMFI:PRCL:PPValue', 'number'), # int or double
    ]
    __slots__ = ()

@utils.register_class
class ParamClip(Clip):
    propertydefs_dict = {}
    class_id = b'PRCL'
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('interp_kind',    'OMFI:PRCL:InterpKind',    'int32'),
        AVBPropertyDef('value_type',     'OMFI:PRCL:ValueType',     'int16'),
        AVBPropertyDef('extrap_kind',    'OMFI:PCRL:ExtrapKind',    'int32'),
        AVBPropertyDef('control_points', 'OMFI:PRCL:ControlPoints', 'list'),
        AVBPropertyDef('fields',         'OMFI:PRCL:Fields',        'int32'),
    ]
    __slots__ = ()

    def read(self, f):
        super(ParamClip, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        self.interp_kind = ctx.read_s32(f)
        self.value_type = ctx.read_s16(f)

        assert self.value_type in (CP_TYPE_INT, CP_TYPE_DOUBLE, CP_TYPE_REFERENCE)

        point_count = ctx.read_s32(f)
        assert point_count >= 0

        self.control_points = []
        for i in range(point_count):
            cp = ParamControlPoint.__new__(ParamControlPoint, root=self.root)
            num = ctx.read_s32(f)
            den = ctx.read_s32(f)
            cp.offset = [num, den]
            cp.timescale = ctx.read_s32(f)

            if self.value_type == CP_TYPE_INT:
                cp.value = ctx.read_s32(f)
            elif self.value_type == CP_TYPE_DOUBLE:
                cp.value = ctx.read_double(f)
            elif self.value_type == CP_TYPE_REFERENCE:
                cp.value = ctx.read_object_ref(self.root, f)
            else:
                raise ValueError("unknown value type: %d" % self.value_type)

            pp_count = ctx.read_s16(f)
            assert pp_count >= 0
            cp.pp = []
            for j in range(pp_count):
                pp = ParamPerPoint.__new__(ParamPerPoint, root=self.root)
                pp.code = ctx.read_s16(f)
                pp.type = ctx.read_s16(f)

                if pp.type == CP_TYPE_DOUBLE:
                    pp.value = ctx.read_double(f)
                elif pp.type == CP_TYPE_INT:
                    pp.value  = ctx.read_s32(f)
                else:
                    raise ValueError("unknown PP type: %d" % pp.type)

                cp.pp.append(pp)

            self.control_points.append(cp)

        for tag in ctx.iter_ext(f):
            if tag == 0x01:
                ctx.read_assert_tag(f, 71)
                self.extrap_kind = ctx.read_s32(f)
            elif tag == 0x02:
                ctx.read_assert_tag(f, 71)
                self.fields = ctx.read_s32(f)
            else:
                raise ValueError("%s: unknown ext tag 0x%02X %d" % (str(self.class_id), tag,tag))

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(ParamClip, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_s32(f, self.interp_kind)
        ctx.write_s16(f, self.value_type)

        ctx.write_s32(f, len(self.control_points))

        for cp in self.control_points:
            ctx.write_s32(f, cp.offset[0])
            ctx.write_s32(f, cp.offset[1])
            ctx.write_s32(f, cp.timescale)

            if self.value_type == CP_TYPE_INT:
                ctx.write_s32(f, cp.value)
            elif self.value_type == CP_TYPE_DOUBLE:
                ctx.write_double(f, cp.value)
            elif self.value_type == CP_TYPE_REFERENCE:
                ctx.write_object_ref(self.root, f, cp.value)
            else:
                raise ValueError("unknown value type: %d" % cp.value_type)

            ctx.write_s16(f, len(cp.pp))
            for pp in cp.pp:

                ctx.write_s16(f, pp.code)
                ctx.write_s16(f, pp.type)

                if pp.type == CP_TYPE_DOUBLE:
                    ctx.write_double(f, pp.value)
                elif pp.type == CP_TYPE_INT:
                    ctx.write_s32(f, pp.value)
                else:
                    raise ValueError("unknown PP type: %d" % pp.type)

        if hasattr(self, 'extrap_kind'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.extrap_kind)

        if hasattr(self, 'fields'):
            ctx.write_u8(f, 0x01)
            ctx.write_u8(f, 0x02)
            ctx.write_u8(f, 71)
            ctx.write_s32(f, self.fields)

        ctx.write_u8(f, 0x03)

@utils.register_helper_class
class ControlPoint(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('offset',     'OMFI:CTRL:Offset',    'rational'),
        AVBPropertyDef('time_scale', 'OMFI:CTRL:TimeScale', 'int32'),
        AVBPropertyDef('value',      'OMFI:CTRL:Value',     'bool'),
        AVBPropertyDef('pp',         'OMFI:CTRL:PP',        'list'),
    ]
    __slots__ = ()

@utils.register_helper_class
class PerPoint(core.AVBObject):
    propertydefs_dict = {}
    propertydefs = [
        AVBPropertyDef('code',    'OMFI:CTRL:PPCode',  'int16'),
        AVBPropertyDef('value',   'OMFI:CTRL:PP',      'rational'),
    ]
    __slots__ = ()


@utils.register_class
class ControlClip(Clip):
    propertydefs_dict = {}
    class_id = b'CTRL'
    propertydefs = Clip.propertydefs + [
        AVBPropertyDef('interp_kind',    'OMFI:CTRL:InterpKin',    'int32'),
        AVBPropertyDef('control_points', 'OMFI:CTRL:ControlPoints', 'list'),
    ]
    __slots__ = ()

    def read(self, f):
        super(ControlClip, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x03)

        self.interp_kind = ctx.read_s32(f)
        count = ctx.read_s32(f)
        self.control_points = []
        # print(self.interp_kind, count)
        #
        # print(peek_data(f).encode("hex"))
        for i in range(count):
            cp = ControlPoint.__new__(ControlPoint, root=self.root)
            a = ctx.read_s32(f)
            b = ctx.read_s32(f)
            cp.offset = [a, b]
            cp.time_scale = ctx.read_s32(f)

            # TODO: find sample with this False
            has_value = ctx.read_bool(f)
            assert has_value == True

            a = ctx.read_s32(f)
            b = ctx.read_s32(f)
            cp.value = [a, b]
            cp.pp = []

            pp_count = ctx.read_s16(f)
            assert pp_count >= 0
            for j in range(pp_count):
                pp = PerPoint.__new__(PerPoint, root=self.root)
                pp.code = ctx.read_s16(f)
                a = ctx.read_s32(f)
                b = ctx.read_s32(f)
                pp.value = [a,b]
                cp.pp.append(pp)

            self.control_points.append(cp)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(ControlClip, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x03)

        ctx.write_s32(f, self.interp_kind)

        ctx.write_s32(f, len(self.control_points))
        for cp in self.control_points:
            ctx.write_s32(f, cp.offset[0])
            ctx.write_s32(f, cp.offset[1])
            ctx.write_s32(f, cp.time_scale)

            ctx.write_bool(f, True)
            ctx.write_s32(f, cp.value[0])
            ctx.write_s32(f, cp.value[1])

            ctx.write_s16(f, len(cp.pp))
            for pp in cp.pp:
                ctx.write_s16(f, pp.code)
                ctx.write_s32(f, pp.value[0])
                ctx.write_s32(f, pp.value[1])

        ctx.write_u8(f, 0x03)

@utils.register_class
class Filler(Clip):
    propertydefs_dict = {}
    class_id = b'FILL'
    __slots__ = ()

    def read(self, f):
        super(Filler, self).read(f)
        ctx = self.root.ictx
        ctx.read_assert_tag(f, 0x02)
        ctx.read_assert_tag(f, 0x01)

        ctx.read_assert_tag(f, 0x03)

    def write(self, f):
        super(Filler, self).write(f)
        ctx = self.root.octx
        ctx.write_u8(f, 0x02)
        ctx.write_u8(f, 0x01)

        ctx.write_u8(f, 0x03)
